//
// C++ Interface: config
//
// Description: 
//
//
// Author: Eric Torstenson <torstenson@chgr.mc.vanderbilt.edu>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef SIMULATIONCONFIG_H
#define SIMULATIONCONFIG_H

#include "defaults.h"

#include "utility/utility.h"
#include "simulation/modelmanager.h"
#include "simulation/basicsample.h"
#include "simulation/poolmanager.h"
#include "simulation/growthrate.h"
#include "simulation/gamodel.h"
#include "simulation/simlamodel.h"
#include "simulation/locselection.h"

namespace GenomeSIM {

using namespace Simulation;
using namespace std;
using namespace Utility;
using namespace Simulation::PopulationGrowth;


typedef map<string, Simulation::Visualization::LocusSelection> LocusSelectionMap;

/**
@brief The configuration parser for genomeSIM

	@author Eric Torstenson <torstenson@chgr.mc.vanderbilt.edu>
*/
class Config : public ConfigurationParser {
public:
    Config();

    ~Config();

	/**
 	 * Dumps a Text Report summarizing the configuration options encountered
	 */
	virtual void GenerateReport(ostream& os);	
	
	/**
	 * @brief Called to assign <i>value</i> to the entry, <i>key</i>
	 * @param key Left hand portion of the configuration line
	 * @param value Right hand portion of the configuration line
	 * @return true/false to indicate that it was a valid entry that was handled
	 */
	virtual bool SetValues(const char *key, const char *value, const char *remainder);

	/**
	 * @brief Check for missing data or impossible configuration settings
	 */
	virtual bool Validate();

	/**
	 * @brief This is called after everything is loaded to aggregate filenames and other composite information
	 */
	bool PostLoad();

	
	/**
	 * @brief This returns the application's pool manager already properly loaded
	 */
	PoolManager *GetPoolManager();

	/**
	 * @brief Returns the growth model described in the configuration file
	 */
	GrowthRate *GetGrowthRate();

	/**
	 * @brief Setup the location mapping function (kosambi/haldane)
	 */
	void SetMappingFn(const char *val);

	/**
	 * @brief Loads the chromosome (loci positions)
	 */
	void LoadChromosome(const char *line);
	
	/**
	 * @brief Loads Initiates the parsing of the configuration file
	 */
	bool LoadSettings(const char *filename);

	/**
	 * @brief Random seed
	 */
	void SetRandomSeed(uint seed);


	void WriteModelToPenetrance(const char *filename);

	/**
	 * @brief returns the filename associated with the growth rate
	 */
	string GetGrowthChartFilename() { return generalSettings.growthChartFilename; }
	
	struct General {
		string outputName;					///<The first part of the reports. This is specified on the cmd line
		uint seed;							///<The random seed 
		uint poolSize;						///<The number interchangeable chromosomes in a given pool
		uint firstGeneration;				///<Which generation did we start with. Used when picking up where we left off
		uint firstDropPoint;				///<The last generation 
		uint dropFrequency;					///<how many generations between drops
		uint dropCount;						///<How many drops are expected be performed
		bool createLDMaps;					///<Do we want to create LD maps using haploview?
		bool writeDetailedLD;				///<Do we want to create Detailed (complete) LD plots?
		bool createSampleLDmaps; 			///<Do we want to generate LD maps for our samples?
		bool keepPhasedOutput;				///<Do we want to keep the phased data after it is done?
		bool doLoad;						///<Are we loading a previous generation? 
		bool closePoolsBetweenAdvancing;	///<Do we keep the pools open between advancements?
		string growthChartFilename;			///<The name of the file containing growth details

		/**
		 * @brief build up the selectors so we can have them ready when the chromosome pools are ready
	 	 */
		LocusSelectionMap locusSelections;
		vector<string> selectionRegions;	///<cache the selections until we can translate them into loci

		General();
		
		/**
	 	 * @brief Add a new locus selection object
	 	 */
		void AddSelector(const char *selLine);
		/**
	 	 * @brief Add a region to search within
	 	 */
		void AddSelectorRegion(const char *selRegion);

		/**
		 * @brief 
		 */
		void ReconcileSelections(PoolManager &pools);

	} generalSettings;

	struct Status {
		//float phenoCopy;					///<fraction of affected that are due to environmental factors
		//float genotypeError;				///<The amount of error in genotypes
		float heteroRate;
		//ModelManager models;
		PenetranceModel *model;
		bool doCreateModel;	
		string modelCfg;
		void SetModel(const char *line);

		PenetranceModel *LoadModel(const char *project, PoolManager *poolMgr);
		PenetranceModel *InitModel(PoolManager *poolmgr, bool requireModel);
		//void AddModel(const char *line);
		void DefineModel(const char *line);
		Status() : heteroRate(0.0), model(NULL) {}
		~Status() { delete model; }
	} statusSettings;

	struct Dataset {
		vector<Sample *> samples;			///<Describe the types of datasets to be used
		uint simSets;						///<The number of datasets per filetype to be written
		uint individualCount;				///<number of indidivuals to be written to a file (this might be interpreted differently for certain 
		bool writeDatasets;
		bool binaryDatasets;
		~Dataset() { 
			uint count = samples.size();
			for (uint i=0; i<count; i++)
				delete samples[i];
		}
		Dataset() : simSets(100), individualCount(100), writeDatasets(false), binaryDatasets(false) { }
	} datasetSettings;						///<Types of datasets

	struct ConfigKeywords {
		//General settings
		static char *RandomSeed;
		static char *PoolSize;
		static char *MaxPoolSize;
		static char *MinPoolSize;
		static char *AlleleLimits;
		static char *SetAlleleFrequency;
		static char *SetAlleleFreqRange;
		static char *GrowthRate;
		static char *DefaultAlleleFreq;
		
		//Status assignment
		static char *PhenoCopy;
		static char *GenotypeError;

		static char *FirstDropPoint;
		static char *DropFrequency;
		static char *DropCount;

		//static char *AddModel;
		static char *SetModel;
		static char *DefineModel;
		static char *LabelBased;
		static char *IndexBased;
		static char *Simla;
		static char *Simpen;
		static char *PenTable;
		static char *GenerateModels;
		static char *IndividualCount;
		static char *FamilyCount;			///<Number of families. This is actually the same as individual count
		static char *DatasetCount;	
		static char *StandardPedigreeHeader;
		
		static char *SetDefaultBlock;
		static char *AddChromosome;			///<Start the definition of a new chromosome
		static char *LoadChromosome;		///<Loads a definition from the chromosome
		static char *SeedChromosome;		///<Creates a chromosome pool based on the seed data
		static char *AddBlock;				///<Add a block to the current chromosome
		static char *RandomizeAlleleFreq;	///<Turns on/off randomization of allele frequencies
		static char *AddDataset;			
		static char *AddFamily;
		static char *PedContinuous;
		static char *AddFamilyRandom;
		static char *AddFamilyType;
		static char *AddCC;
		static char *AddCont;
		static char *AddContTails;
		static char *OffspringPerMating;	///<Keyword to set min/max children per mating (in advancement)

		static char *GenerateOverviewLD;
		static char *MappingFN;
		static char *BlockReportSize;
		static char *MaxSnpsPerRow;
		static char *MaxSNPDistance;		///<For LD Comparisons

		static char *GAFitnessThreshold;
		static char *GAIterations;

		static char *MinAlleleFreqThresh;
		static char *IncludeConfiguration;

		static char *DoDumpGenerationZero;

		static char *TrueTypeFontName;		///<The font to be used
		static char *BlockReportBuffer;		///<The number of snps to be used on either side of a block in an LD plot
		static char *CssFilename;			///<Filename to be used as the CSS style sheet. This will be written to all HTML report pages and must be available in the location specified relative to the report whenever it is viewed
		static char *ClosePoolsBetweenAdvancements;

		static char *UseOriginalCrossing;
		static char *PedUseOriginalCross;

		static char *UseAltLD;
		static char *DrawRSquared;
		static char *DrawDPrime;
		static char *WriteLDReport;
		//These two parameters allow for limiting the LD calculations in order to perform
		//scans in reasonable amounts of time
		static char *MaxLdIndividualCount;	///<Number of individuals 
		static char *PlotScanSize;			///<Number of snps

		static char *BinaryDatasets;

		static char *LocSelection;			///<Add a locus selection object
		static char *AddRegion;				///<Add a region to a selection object

		static char *SimultaneousChroms;	///<Number of chroms to be worked on at once
		static char *ThreadsPerChrom;		///<Number of threads to be launched per chrom

		static char *MaxThreadCount;		///<Specify the number of threads to be used
		static char *OverideRecThreadCount;

		static char *NoFastLD;				///<Turn off sampled LD
		static char *TargetPopSize;			///<Pop. Size where GS will stop once it is reached
		static char *LociPerChromReported;	///<Max Number of loci / chromosomes to be reported

		static char *UseAdamEve;
		static char *UseEden;
		static char *FounderCount;
		static char *FounderDistortion;
		static char *MaxRepeatCount;
		static char *ParentDistortion;
		static char *ChildDistortion;


		
	};
	//string outputPrefix;					///<This is used in the generation of each filename
	string errorMsg;						///<Used to report any parsing errors;

	string GetConfigurationFilename() { return configurationFilename; }
	void SetConfigurationFilename(const char *filename) { configurationFilename = filename; }
protected:
	PoolManager pools;
	ChromPool::BlockDefinition defaultBlock;///<Used to define/store the default block qualities
	ChromPool *currPool;					///<Just used to keep up with the current pool. It should already be in the manager


	string GetRemainderString(const char *line);
	void SetAlleleFrequency(const char *line);
	void SetAlleleFreqRange(const char *line);
	void AddChromosome(const char *line);
	void SeedChromosome(const char *line);
	void AddBlock(const char *line);
	void SetDefaultBlock(const char *line);
	void SetGrowthRate(const char *line);
	void SetDefaultAlleleFreq(const char *line);
	void AddDataset(const char *line);
	
	bool doDumpGenerationZero;

	
	GrowthRate *growthModel;
	bool growthModelConfigured;

	string configurationFilename;
};

					

inline
void Config::Status::DefineModel(const char *line) {
	modelCfg = line;



}

inline
void Config::General::AddSelectorRegion( const char *selLine) {
	selectionRegions.push_back(selLine);
}

inline
Config::General::General() : outputName(""), seed(DEFAULT_SEED), poolSize(DEFAULT_POOL_SIZE), 
		firstGeneration(0), firstDropPoint(1000), dropFrequency(250), dropCount(4), 
		createLDMaps(false), writeDetailedLD(false), createSampleLDmaps(false), 
		keepPhasedOutput(true), doLoad(false), closePoolsBetweenAdvancing(false) {

	//This default selection is just a place holder for "global" regions
	locusSelections["ALL_SELECTORS"] = Simulation::Visualization::LocusSelection(0.5, 0.0, 0.5, 10, 0, 50);
	locusSelections["ALL_SELECTORS"].SetLabel("ALL_SELECTORS");
}


/*
inline
void Config::Status::SetModel(const char *line) {
	stringstream ss(line);
	string filename = "";
	float percentage = 0.0;
	string cmd;
	
	ss>>cmd>>filename>>percentage;

	if (model)  {
		cout<<"Overwriting previous model definition. (";
		model->GenerateReport(cout, 0);
		cout<<")\n";
		delete model;
	}

	if (percentage > 0.0) {
		model = new PenetranceModel();
		model->SetFilename(filename.c_str());
	}
}*/

/*
inline
void Config::Status::AddModel(const char *line) {
	stringstream ss(line);
	string filename = "";
	float percentage = 0.0;
	string cmd;

	ss>>cmd>>filename>>percentage;

	if (percentage > 0.0 ) { 
		PenetranceModel *newModel = new PenetranceModel(models.GetModelCount(), percentage);
		newModel->SetFilename(filename.c_str());
		//newModel->Load();
		models.AddModel( newModel);
	}
	else {
		cout<<"Misconfigured model definition: \n\t"<<line<<"\n";
		cout<<"Correct syntax: \tADD_MODEL filename probability\n";
		abort();
	}
}
*/
}

#endif
