//
// C++ Implementation: individual
//
// Description: 
//
//
// Author: Eric Torstenson <torstenson@chgr.mc.vanderbilt.edu>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include <sstream>
#include "individual.h"
#include "penetrancemodel.h"
namespace Simulation {

using namespace std;
using namespace Utility;
using namespace StatusModel;

/**
 * @brief Standard pedigree is treated as the 10 column version
 */
bool Individual::StandardPedigreeHeader 	= false;

/**
 * @brief If an allele's minor allele frequency is lower than this, we won't write it to datasets 
 */
float Individual::minAlFreqThreshold		= 0.00;

Individual::Individual(uint id, uint pedID, uint chrCount /*=1*/) : id(id), pedID(pedID), patID(0), 
			matID(0), includeInDataset(true), status(0), chromosomeCount(chrCount), mom(NULL), 
			dad(NULL), chrID1(chrCount), chrID2(chrCount) {
	matChrom = new Chromosome[chrCount];
	patChrom = new Chromosome[chrCount];
	genePoolIDs = new string[chrCount];
	missingData = new BitSetType[chrCount];
}

Individual::Individual() : id(0), pedID(0), patID(0), matID(0), includeInDataset(true), status(0),
		 matChrom(NULL), patChrom(NULL), chromosomeCount(0), genePoolIDs(NULL), mom(NULL), 
		dad(NULL), chrID1(0), chrID2(0) { }

void Individual::Init(uint id, uint pedID, uint chrCount /*=1*/)  {
	id = id;
	pedID = pedID;
	chromosomeCount = chrCount;
	if (matChrom) 
		matChrom = new Chromosome[chrCount];

	if (patChrom)
		patChrom = new Chromosome[chrCount];
	if (genePoolIDs)
		delete[] genePoolIDs;
	genePoolIDs = new string[chrCount];
	chrID1.resize(chrCount);
	chrID2.resize(chrCount);
}

bool Individual::DoIncludeInDataset() {
	return includeInDataset;
}

void Individual::DoIncludeInDataset( bool doinclude){
	includeInDataset=doinclude;
}

float Individual::MinorAlleleFreq(uint chID, uint locID) {
	float maf = 0.0;				///<Minor allele freq
	if (patChrom) {
		Locus l = patChrom[chID].GetLocus(locID);
		maf = l.GetMinAlleleFreq();
	}
	return maf;		
}

Individual::~Individual()	{
	if (matChrom)
		delete[] matChrom;
	if (patChrom)
		delete[] patChrom;
	if (genePoolIDs)
		delete[] genePoolIDs;
	if (missingData)
		delete[] missingData;
}




void Individual::SetDLPedigreeMeta(Individual *dad, Individual *mom) {
	this->patID=dad->GetID();
	this->matID=mom->GetID();
	this->dad=dad;
	this->mom=mom;
	SetChromosomalData(dad->DLCross(), mom->DLCross());
}

void Individual::SetPedigreeMeta(Individual *dad, Individual *mom) {
	this->patID=dad->GetID();
	this->matID=mom->GetID();
	this->dad=dad;
	this->mom=mom;
	SetChromosomalData(dad->Cross(), mom->Cross());
}

void Individual::SetPoolID(uint chromeID, const char *id) {
	assert(chromeID < chromosomeCount && genePoolIDs);
	genePoolIDs[chromeID]=id;
}

std::string Individual::GetPoolID(uint chromID) {
	assert(chromID <chromosomeCount);
	return genePoolIDs[chromID];
}

void Individual::SetChromosomalData(Chromosome *p, Chromosome *m) {
	if (matChrom)
		delete[] matChrom;
	if (patChrom)
		delete[] patChrom;

	matChrom = m;
	patChrom = p;

	for (uint i=0; i<chromosomeCount; i++)
		missingData[i].resize(p[i].LociCount(), false);
}

void Individual::SetChromIDs(uint chromID, uint chID1, uint chID2) {
	chrID1[chromID] = chID1; 
	chrID2[chromID] = chID2;
}

bool Individual::NeedsCompleteGenotypeInformation() {
	return mom == NULL;
}

void Individual::SetChromosomalData(uint chromID, Chromosome &p, Chromosome &m) {
	assert(chromID < chromosomeCount);
	matChrom[chromID] = m;
	patChrom[chromID] = p;
	missingData[chromID].resize(p.LociCount(), false);
	//cout<<"Setting Missing ("<<chromID<<") to "<<p.LociCount()<<"\n";
}


uint Individual::GetStatus() { 
	return status;
}

void Individual::SetStatus(int isAffected) {
	status = isAffected;
}

bool Individual::IsAffected() {
	return status != 0;
}


void Individual::SetOutcome(float& outcome) {
	this->outcome=outcome;
}
float Individual::GetOutcome() {
	return outcome;
}

bool Individual::MissingGenotype(uint chrID, uint locus) {
	return missingData[chrID][locus];
}
bool Individual::ClearLocus( uint chrID, uint locus) {
	assert(chromosomeCount > chrID);
	if (missingData[chrID].size() <= locus)
		cout<<"Hm, we are about to die! Missing data["<<chrID<<"].size() == "<<missingData[chrID].size()<<" & Locus == "<<locus<<"\n";
	assert(missingData[chrID].size() > locus);

	if (missingData[chrID][locus]) 
		return false;
	else 
		missingData[chrID][locus] = true;
	return true;
}

/**
 * @Note errorDir == 0  -Allow for randomized direction
 */
bool Individual::ChangeGenotype(uint chrID, uint locus, int errorDir) {

	if (errorDir == 0)  {
		if (Utility::Random::globalGenerator((float)1.0) > 0.5)
			errorDir = 1;
		else
			errorDir = -1;
	}


	int gt = GetGenotype(chrID, locus) + errorDir;

	bool success=true;
	switch (gt) {
		case 0: 
			matChrom[chrID][locus] = false;
			patChrom[chrID][locus] = false;
			break;
		case 1:
			if (errorDir > 0) {
				if (Utility::Random::globalGenerator((float)1.0) > 0.5) 
					matChrom[chrID][locus]=true;
				else
					patChrom[chrID][locus]=true;
			}
			else {
				if (Utility::Random::globalGenerator((float)1.0) > 0.5) 
					matChrom[chrID][locus]=false;
				else
					patChrom[chrID][locus]=false;
			}
			break;
		case 2:
			matChrom[chrID][locus]=true;
			matChrom[chrID][locus]=true;
			break;
		default:
			success=false;
	}
	return success;
}



int Individual::ApplyStatus(DiseaseModel *model) {
	int isAffected = 0;
	if (model) {
		size_t modelSize = model->GetModelSize();
		vector<uint> genotypes;
		for (uint i=0; i<modelSize; i++)  {
			StatusModel::DiseaseLocus l = model->GetLocus(i);
			genotypes.push_back(GetGenotype(l.chromosome, l.locusIdx));
		}

		//For continuous data, we need to know if there is any relevance here. If so, 
		//How do we determine the status relationship? And if it is irrelavent, how
		//do we show that it doesn't matter?
		isAffected = model->GetStatus(genotypes, outcome);
	}
	else {
		if (Utility::Random::globalGenerator.drand() < 0.5) 
			isAffected = 1;
		outcome = 0.0;
	}
	SetStatus(isAffected);

	return isAffected;	
}



Individual *Individual::DLCross(Individual *father, size_t id) {
	Individual *child = new Individual(id, father->GetPedigreeID(), chromosomeCount);
	child->SetDLPedigreeMeta(father, this);
	return child;
}

Individual *Individual::Cross(Individual *father, size_t id) {
//	Chromosome *pDNA = father->Cross();
//	Chromosome *mDNA = Cross();
	Individual *child = new Individual(id, father->GetPedigreeID(), chromosomeCount);
	child->SetPedigreeMeta(father, this);
//	child->SetChromosomalData(pDNA, mDNA);
	return child;
}

Chromosome *Individual::DLCross() {
	Chromosome *newChr = new Chromosome[chromosomeCount];
	size_t eventCount;
	for (uint i=0; i<chromosomeCount; i++)
		newChr[i] = patChrom[i].DLCross(matChrom[i], eventCount);

	return newChr;
}

Chromosome *Individual::Cross() {
	Chromosome *newChr = new Chromosome[chromosomeCount];
	size_t eventCount;						///<need to capture the number of x-over events
	for (uint i=0; i<chromosomeCount; i++)
		newChr[i] = patChrom[i].Cross(matChrom[i], eventCount);

	return newChr;
}

int Individual::GetGenotype(uint chrID, uint locus) {

	if (chromosomeCount <= chrID || locus >= matChrom[chrID].LociCount() || locus >= patChrom[chrID].LociCount()) {
		cout<<"-------------------------- "<<chrID<<" < "<<chromosomeCount<<" && "<<matChrom[chrID].LociCount()<<" < "<<locus<<" > "<<patChrom[chrID].LociCount()<<"\n";
		cout<<"--------------------------We asked for an invalid memory location! (individual.cpp:196)\n";
		assert(0);
	}
	return matChrom[chrID][locus] + patChrom[chrID][locus];
}
/*
uint Individual::GetChromosomeCount() {
	return chromosomeCount;
}

size_t Individual::GetSnpCount(uint chID) {
	return patChrom[chID].LociCount;
}*/
int Individual::GetSnpCount() {
	uint lociCount = 0;
	for (uint chID=0; chID<chromosomeCount; chID++) 
		lociCount+=patChrom[chID].LociCount();

	return lociCount;
}

void Individual::ExpandModelLoci(uint chID, vector<uint>& modelLoci) {
	matChrom[chID].ExpandModelLoci(modelLoci);
	patChrom[chID].ExpandModelLoci(modelLoci);


}

void Individual::ResolveGenotypes(vector<uint> &modelLoci) {
	
	for (uint chID=0; chID<chromosomeCount; chID++) {
		ResolveGenotypes(chID, modelLoci);
	}
}

void Individual::ResolveGenotypes(uint chID, vector<uint> &modelLoci) {
		vector<uint> genotypes(modelLoci.size());
		for (uint i=0; i<modelLoci.size(); i++) 
			genotypes[i] = GetGenotype(chID, modelLoci[i]);

		stringstream ss;

		patChrom[chID].ResolveGenotypes();
		matChrom[chID].ResolveGenotypes();	
}


void Individual::WritePedigree(std::ostream& meta, std::ostream&genotypes, bool writeMissingData, bool doWriteStatus /*=true*/) {
	static int statusConvertor[] = {1, 2, 0};
	meta<<pedID<<" ";								//Column #1
	meta<<id<<" ";								//Column #2
	meta<<matID<<" ";								//Column #3
	meta<<patID<<" ";								//Column #4

	if (StandardPedigreeHeader) {
		meta<<"0 ";			//uint nextOffspring;	//Column #5
		meta<<"0 ";			//uint nextPatSib;		//Column #6
		meta<<"0 ";			//uint nextMatSib;		//Column #7
	}
	if (id<3)
		meta<<id<<" ";
	else
		meta<<"0 ";			//uint sex;				//Column #8
	if (StandardPedigreeHeader) {
		meta<<"0 ";			//uint probStat;		//Column #9
	}
	meta << statusConvertor[status] << " ";

	for (uint chID=0; chID<chromosomeCount; chID++) 
		patChrom[chID].WriteBinary(genotypes);

	for (uint chID=0; chID<chromosomeCount; chID++) 
		matChrom[chID].WriteBinary(genotypes);
	
	if (writeMissingData) {
		for (uint chID=0; chID<chromosomeCount; chID++) {
			BitSetType &misses = missingData[chID];
			uint totalBlocks = (misses.size() / boost::dynamic_bitset<>::bits_per_block) + 1;
			dynamic_bitset<>::block_type raw[totalBlocks];

			//Convert the bitset to raw data
			to_block_range(misses, raw);

			genotypes.write((char*)&raw, (sizeof(dynamic_bitset<>::block_type)*totalBlocks));	
		}
	}

}

void Individual::WritePedigree(std::ostream& os, uint *genotypeCounts, bool writeOutcome) {
	static int statusConvertor[] = {1, 2, 0};
	os<<pedID<<" ";								//Column #1
	os<<id<<" ";								//Column #2
	os<<matID<<" ";								//Column #3
	os<<patID<<" ";								//Column #4

	if (StandardPedigreeHeader) {
		os<<"0 ";			//uint nextOffspring;	//Column #5
		os<<"0 ";			//uint nextPatSib;		//Column #6
		os<<"0 ";			//uint nextMatSib;		//Column #7
	}
	if (id<3)
		os<<id<<" ";
	else
		os<<"0 ";			//uint sex;				//Column #8
	if (StandardPedigreeHeader) {
		os<<"0 ";			//uint probStat;		//Column #9
	}
	os << statusConvertor[status] << " ";
	if (writeOutcome)
		os<<outcome<<" ";

	int locPos = 0;
	//Write each of the chromosomes to the line
	for (uint chID=0; chID<chromosomeCount; chID++) {

		uint lociCount=patChrom[chID].LociCount();
		os << " ";
		for (uint locID=0; locID<lociCount; locID++) {
			//Let's make sure the minor allele frequencies are within the threshold
			if (MinorAlleleFreq(chID, locID) >= (minAlFreqThreshold - 0.0001)) {
				//If this is set as missing, return -1
				if (missingData[chID][locID]) {
					if (genotypeCounts)
						genotypeCounts[locPos++ * 4]++;
					os << "0 0 ";
				}
				else {
					int gt = GetGenotype(chID, locID);
					if (genotypeCounts)
						genotypeCounts[locPos++ * 4 + gt+1]++;
					switch(gt){
						case 0:
							os << "1 1 ";
							break;
						case 1:
							os << "1 2 ";
							break;
						case 2:
							os << "2 2 ";
							break;
					}
				}
			}	
			else 
				cout<<"Skipping locus "<<locID + 1<<" frequency "<<MinorAlleleFreq(chID, locID)<<"\n";
		}
	}
	
  os << std::endl;
}

void Individual::WriteContinuous(std::ostream&os, uint *genotypeCounts, bool doWriteStatus /*=true*/) {

	if (doWriteStatus)
		os << outcome << " ";

	int locPos = 0;
	for (uint chID=0; chID<chromosomeCount; chID++) {
		uint lociCount=patChrom[chID].LociCount();
		os << " ";
		for (uint locID=0; locID<lociCount; locID++) {
			//Let's make sure the minor allele frequencies are within the threshold
			if (MinorAlleleFreq(chID, locID) >= (minAlFreqThreshold - 0.0001)) {
	
				if (missingData[chID][locID]) {
					if (genotypeCounts)
						genotypeCounts[locPos++ * 4]++;
					os <<" -1";
				}			//If this is set as missing, return -1
				else {
					int gt=GetGenotype(chID, locID);
					if (genotypeCounts)
						genotypeCounts[locPos * 4 + gt + 1]++;
					os <<" "<<gt;
				}
			}
		}
	}
  os << std::endl;
}

void Individual::WriteMDR(std::ostream&os, uint *genotypeCounts, bool doWriteStatus /*=true*/) {

	if (doWriteStatus)
		os << status << " ";

	int locPos = 0;
	for (uint chID=0; chID<chromosomeCount; chID++) {
		uint lociCount=patChrom[chID].LociCount();
		os << " ";
		for (uint locID=0; locID<lociCount; locID++) {
			//Let's make sure the minor allele frequencies are within the threshold
			if (MinorAlleleFreq(chID, locID) >= (minAlFreqThreshold - 0.0001)) {
	
				if (missingData[chID][locID]) {
					if (genotypeCounts)
						genotypeCounts[locPos++ * 4]++;
					os <<" -1";
				}			//If this is set as missing, return -1
				else {
					int gt=GetGenotype(chID, locID);
					if (genotypeCounts)
						genotypeCounts[locPos * 4 + gt + 1]++;
					os <<" "<<gt;
				}
			}
		}
	}
  os << std::endl;
}

void Individual::ReadBinaryMdr(std::ifstream *genotypes, bool readMissingData, bool doWriteStatus, vector<uint> &chrSize) {
	//if (doWriteStatus)
	genotypes->read((char *)&status, 4);
		
	for (uint chID=0; chID<chromosomeCount; chID++) 
		patChrom[chID].ReadBinary(*genotypes, chrSize[chID], NULL, true);

	for (uint chID=0; chID<chromosomeCount; chID++) 
		matChrom[chID].ReadBinary(*genotypes, chrSize[chID], NULL, true);
	
	if (readMissingData) {
		for (uint chID=0; chID<chromosomeCount; chID++) {
			uint totalBlocks = (chrSize[chID] / boost::dynamic_bitset<>::bits_per_block) + 1;
			dynamic_bitset<>::block_type raw[totalBlocks];
			genotypes->read((char *)raw[0], (sizeof(dynamic_bitset<>::block_type)*totalBlocks));
			missingData[chID] = dynamic_bitset<>(&raw[0], &raw[totalBlocks]);
			missingData[chID].resize(chrSize[chID]);
		}
	}
	
}
void Individual::WriteContinuous(std::ostream& meta, std::ostream&genotypes, bool writeMissingData, bool doWriteStatus /*=true*/) {

	//For now, we don't have a binary convention to leave off the status
	//if (doWriteStatus)
		genotypes.write((const char*)&outcome, 4);
		
	for (uint chID=0; chID<chromosomeCount; chID++) 
		patChrom[chID].WriteBinary(genotypes);

	for (uint chID=0; chID<chromosomeCount; chID++) 
		matChrom[chID].WriteBinary(genotypes);
	
	if (writeMissingData) {
		for (uint chID=0; chID<chromosomeCount; chID++) {
			BitSetType &misses = missingData[chID];
			uint totalBlocks = (misses.size() / boost::dynamic_bitset<>::bits_per_block) + 1;
			dynamic_bitset<>::block_type raw[totalBlocks];

			//Convert the bitset to raw data
			to_block_range(misses, raw);

			genotypes.write((char*)&raw, (sizeof(dynamic_bitset<>::block_type)*totalBlocks));	
		}
	}
}

void Individual::WriteMDR(std::ostream& meta, std::ostream&genotypes, bool writeMissingData, bool doWriteStatus /*=true*/) {

	//For now, we don't have a binary convention to leave off the status
	//if (doWriteStatus)
		genotypes.write((const char*)&status, 4);
		
	for (uint chID=0; chID<chromosomeCount; chID++) 
		patChrom[chID].WriteBinary(genotypes);

	for (uint chID=0; chID<chromosomeCount; chID++) 
		matChrom[chID].WriteBinary(genotypes);
	
	if (writeMissingData) {
		for (uint chID=0; chID<chromosomeCount; chID++) {
			BitSetType &misses = missingData[chID];
			uint totalBlocks = (misses.size() / boost::dynamic_bitset<>::bits_per_block) + 1;
			dynamic_bitset<>::block_type raw[totalBlocks];

			//Convert the bitset to raw data
			to_block_range(misses, raw);

			genotypes.write((char*)&raw, (sizeof(dynamic_bitset<>::block_type)*totalBlocks));	
		}
	}
}

void Individual::WriteChromosome(std::ostream& os, uint chrID) {
	patChrom[chrID].ShowGenotypes(100, '\t');
	matChrom[chrID].ShowGenotypes(100, '\t');
}

/**
 * I'm not completely certain this is correct. I need to visit the haploview manual for a refresher
 */
void Individual::WritePhased(std::ostream& os, uint indID, bool useFreqThreshold) {
	std::stringstream s1, s2;

	s1<<indID<<" ";				//Column #1
	s2<<indID<<" ";				//Column #1

	for (uint chID=0; chID<chromosomeCount; chID++) {
		uint lociCount=patChrom[chID].LociCount();
		os << " ";
		for (uint locID=0; locID<lociCount; locID++) {
			if (MinorAlleleFreq(chID, locID) >= (minAlFreqThreshold - 0.0001)) {
				s1<<patChrom[chID].At(locID) + 1<<" ";
				s2<<matChrom[chID].At(locID) + 1<<" ";
			}
		}
	}
	
 	os << s1.str() << std::endl;
 	os << s2.str() << std::endl;

}


}
